%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:-import new_trie/1,
        trie_intern/2,
        trie_intern/4,
        trie_interned/2,
        trie_interned/4 from intern.
:-import member/2,
        length/2,
        append/3,
        select/3 from basics.

:- ['tr_utils.P'].

% dynamic initial_state/1 predicate: "initial_state(FactId)"
:- dynamic(initial_state/1).

% dynamic insDelsRoot/1 predicate containing the root of the trie that
%  contains materialized state logs in the format:
%  "ins/dels(ListOfAll_FactIds_Ins/Dels_Ordered,QueryIns/DelsRoot)"
:- dynamic(insDelsRoot/1).

% dynamic stateRoot/1 predicate contains the root of the trie that contains 
%  the states: "state(InsLeaf,DelsLeaf)"
:- dynamic(stateRoot/1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initializeCTRSystem(+CTRProgram): initializes the transactional system
initializeCTRSystem(CTRProgram,State0):-
        retractall(dynamic_predicate(_,_)),
        retractall(tr_database(_)),
        retractall(tr_clause(_,_)),
        load_dyn(CTRProgram),
        findall(Head,tr_database(Head),ListFacts),
        no_backtracking_facts_to_factids(ListFacts,ListFactIds),
        retractall(initial_state(_)),
        retractall(insDelsRoot(_)),
        retractall(stateRoot(_)),
        flush_initial_state(ListFactIds),
        new_trie(InsDelsRoot),
        assert(insDelsRoot(InsDelsRoot)),
        new_trie(StateRoot),
        assert(stateRoot(StateRoot)),
        trie_intern([],InsDelsRoot,InsLeaf,_),
        trie_intern([],InsDelsRoot,DelsLeaf,_),
        trie_intern(state(InsLeaf,DelsLeaf),StateRoot,State0,_),
        (
        	tr_debug(on)
        	-> inc_counter(state_copies,_) % the initial state is created
        	; true
        ),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% query_tries
:- dynamic(log_query_tries/2).
:- dynamic(state_query_tries/3).

check_make_query_tries(S1,QueryInsRoot,QueryDelsRoot):-
        state_query_tries(S1,QueryInsRoot,QueryDelsRoot),
        !.
check_make_query_tries(S1,QueryInsRoot,QueryDelsRoot):-
        stateRoot(StateRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
        insDelsRoot(InsDelsRoot),
        (
        	log_query_tries(InsLeaf,QueryInsRoot)
        	-> true
        	; (
        		trie_interned(LI1,InsDelsRoot,InsLeaf,_),
        		list_to_trie(LI1,QueryInsRoot),
        		assert(log_query_tries(InsLeaf,QueryInsRoot))
        	)
        ),
        (
        	log_query_tries(DelsLeaf,QueryDelsRoot)
        	-> true
        	; (
        		trie_interned(LD1,InsDelsRoot,DelsLeaf,_),
        		list_to_trie(LD1,QueryDelsRoot),
        		assert(log_query_tries(DelsLeaf,QueryDelsRoot))
        	)
        ),	
        assert(state_query_tries(S1,QueryInsRoot,QueryDelsRoot)),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% keeps the list of initial fact ids (for debugging purposes)
:- dynamic(initial_state_list/1).

% flush_initial_state/1
flush_initial_state(ListFactIds):-
	retractall(initial_state_list(_)),
	sort(ListFactIds,OrdListFactIds),
	assert(initial_state_list(OrdListFactIds)),
	create_initialState(ListFactIds).

% create_initialState/1
% create_initialState(+ListFactIds)
create_initialState([]).
create_initialState([H|T]):-
        assert(initial_state(H)),
        create_initialState(T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check_store_tabled_state(+S1,-S1t): checks/puts the current state into the
%  tabled states
check_store_tabled_state(S1,S1t):-
        S1 = S1t.

absolute_state(S1,S1t):-
	S1 = S1t,
	!.

differential_state(S1,S1t):-
        S1 = S1t.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_select(+Fact,+State): checks if a fact is in the state
db_select(Fact,State):-
        check_make_query_tries(State,QueryInsRoot,QueryDelsRoot),
        fact_to_factid(Fact,FactId),
        db_select_internal(FactId,QueryInsRoot,QueryDelsRoot).

% db_select_internal/3
db_select_internal(FactId,_QueryInsRoot,QueryDelsRoot):-
        initial_state(FactId), % initial state and it was not deleted
        \+( trie_interned(FactId,QueryDelsRoot)),
        !.
db_select_internal(FactId,QueryInsRoot,_QueryDelsRoot):-
        % it was inserted
        trie_interned(FactId,QueryInsRoot),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_ins(+Fact,+S1,-S2)
db_ins(Fact,S1,S2):-
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        no_backtracking_fact_to_factid(Fact,FactId),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
        check_make_query_tries(S1,QueryInsRoot,QueryDelsRoot),
        trie_interned(Ins,InsDelsRoot,InsLeaf,_),
        trie_interned(Dels,InsDelsRoot,DelsLeaf,_),
        db_ins_internal(FactId,S1,S2,StateRoot,InsDelsRoot,
        	InsLeaf,QueryInsRoot,DelsLeaf,QueryDelsRoot,Ins,Dels),
        !.

% db_ins_internal/13
db_ins_internal(FactId,_S1,S2,StateRoot,InsDelsRoot,InsLeaf,_QueryInsRoot,
                _DelsLeaf,_QueryDelsRoot,_Ins,Dels):-
        % re-insertion of a deleted fact
        del_ord(FactId,Dels,NewDels),
        (
          ( % the set NewDels was seen before
            trie_interned(NewDels,InsDelsRoot,NewDelsLeaf,_)
          );( % the set NewDels was not seen before
            % save the NewDels
            trie_intern(NewDels,InsDelsRoot,NewDelsLeaf,_)
          )
        ),
        % save/check the state
	(
        	tr_debug(on)
        	-> inc_counter(state_comps,_)
        	; true
        ),
        (
		(tr_debug(on),
		\+(trie_interned(state(InsLeaf,NewDelsLeaf),StateRoot)))
		-> inc_counter(state_copies,_)
		; true
	),
        trie_intern(state(InsLeaf,NewDelsLeaf),StateRoot,S2,_),
        !.
db_ins_internal(FactId,S1,S1,_,_,_,_,_,_,_,_):-
        % non-strict updates: the fact was in the initial state,
        %       but it was never deleted
        initial_state(FactId),
        !.
db_ins_internal(FactId,S1,S1,_,_,_,QueryInsRoot,_,_,_,_):-
        % non-strict updates; the state was inserted in the current state
        trie_interned(FactId,QueryInsRoot),
        !.
db_ins_internal(FactId,_S1,S2,StateRoot,InsDelsRoot,_InsLeaf,_QueryInsRoot,
                DelsLeaf,_QueryDelsRoot,Ins,_Dels):-
        % insertion of a new fact (not in the initial state):
        %\+(initial_state(Pid))
        ins_ord(FactId,Ins,NewIns),
        (
          ( % the set NewIns was seen before
            trie_interned(NewIns,InsDelsRoot,NewInsLeaf,_)
          );( % the set NewIns was not seen before
            trie_intern(NewIns,InsDelsRoot,NewInsLeaf,_)
          )
        ),
	(
        	tr_debug(on)
        	-> inc_counter(state_comps,_)
        	; true
        ),
        (
		(tr_debug(on),
		\+(trie_interned(state(NewInsLeaf,DelsLeaf),StateRoot)))
		-> inc_counter(state_copies,_)
		; true
	),
        trie_intern(state(NewInsLeaf,DelsLeaf),StateRoot,S2,_),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_del(+Fact,+S1,-S2)
db_del(Fact,S1,S2):-
        no_backtracking_fact_to_factid(Fact,FactId),
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
        check_make_query_tries(S1,QueryInsRoot,QueryDelsRoot),
        trie_interned(Ins,InsDelsRoot,InsLeaf,_),
        trie_interned(Dels,InsDelsRoot,DelsLeaf,_),
        db_del_internal(FactId,S1,S2,StateRoot,InsDelsRoot,
        	InsLeaf,QueryInsRoot,DelsLeaf,QueryDelsRoot,Ins,Dels),
        !.

% db_del_internal/13
db_del_internal(FactId,_S1,S2,StateRoot,InsDelsRoot,_InsLeaf,_QueryInsRoot,
                DelsLeaf,_QueryDelsRoot,Ins,_Dels):-
        % deletion of an already inserted fact
        del_ord(FactId,Ins,NewIns),
        (
          ( % the set NewIns was seen before
            trie_interned(NewIns,InsDelsRoot,NewInsLeaf,_)
          );( % the set NewIns was not seen before
            trie_intern(NewIns,InsDelsRoot,NewInsLeaf,_)
          )
        ),
        (
        	tr_debug(on)
        	-> inc_counter(state_comps,_)
        	; true
        ),
        (
		(tr_debug(on),
		\+(trie_interned(state(NewInsLeaf,DelsLeaf),StateRoot)))
		-> inc_counter(state_copies,_)
		; true
	),
        trie_intern(state(NewInsLeaf,DelsLeaf),StateRoot,S2,_),
        !.
db_del_internal(FactId,S1,S1,_,_,_,_,_,_,_,_):-
        % non-strict updates: the fact was not in the initial state,
        %       but it was never inserted either
        \+( initial_state(FactId) ),
        !.
db_del_internal(FactId,S1,S1,_,_,_,_,_,QueryDelsRoot,_,_):-
        % non-strict updates: the fact was already deleted in the state
        trie_interned(FactId,QueryDelsRoot),
        !.
db_del_internal(FactId,_S1,S2,StateRoot,InsDelsRoot,InsLeaf,_QueryInsRoot,
                _DelsLeaf,_QueryDelsRoot,_Ins,Dels):-
        % deletion of a fact
        ins_ord(FactId,Dels,NewDels),
        (
          ( % the set NewDels was seen before
            trie_interned(NewDels,InsDelsRoot,NewDelsLeaf,_)
          );( % the set NewDels was not seen before
            trie_intern(NewDels,InsDelsRoot,NewDelsLeaf,_)
          )
        ),
        (
        	tr_debug(on)
        	-> inc_counter(state_comps,_)
        	; true
        ),
        (
		(tr_debug(on),
		\+(trie_interned(state(InsLeaf,NewDelsLeaf),StateRoot)))
		-> inc_counter(state_copies,_)
		; true
	),
        trie_intern(state(InsLeaf,NewDelsLeaf),StateRoot,S2,_),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
