%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:-import new_trie/1,
        trie_intern/2,
        trie_intern/4,
        trie_interned/2,
        trie_interned/4 from intern.

:-import member/2,
        length/2,
        append/3,
        select/3 from basics.

:- import sort/2 from setof.

:- ['tr_utils.P'].

% dynamic initial_state/1 predicate: "initial_state(FactId)"
:- dynamic(initial_state/1).

% dynamic insDelsRoot/1 predicate containing the root of the trie that
%  contains materialized state logs in the format:
%  "ins/dels(Size,ListOfAll_FactIds_Ins/Dels_Ordered)"
:- dynamic(insDelsRoot/1).

% dynamic stateRoot/1 predicate contains the root of the trie that contains 
%  the states: "state(InsLeaf,DelsLeaf)"
:- dynamic(stateRoot/1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initializeCTRSystem(+CTRProgram): initializes the transactional system
initializeCTRSystem(CTRProgram,State0):-
        retractall(dynamic_predicate(_,_)),
        retractall(tr_database(_)),
        retractall(tr_clause(_,_)),
        load_dyn(CTRProgram),
        findall(Head,tr_database(Head),ListFacts),
        no_backtracking_facts_to_factids(ListFacts,ListFactIds),
        retractall(initial_state(_)),
        retractall(insDelsRoot(_)),
        retractall(stateRoot(_)),
        flush_initial_state(ListFactIds),
        new_trie(InsDelsRoot),
        assert(insDelsRoot(InsDelsRoot)),
        new_trie(StateRoot),
        assert(stateRoot(StateRoot)),
        % saving the first state
        trie_intern([],InsDelsRoot,InsLeaf,_),
        trie_intern([],InsDelsRoot,DelsLeaf,_),
        trie_intern(state(InsLeaf,DelsLeaf),StateRoot,RelativeState0,_),
	(
        	tr_debug(on)
        	-> inc_counter(state_copies,_) % the initial state is created
        	; true
        ),
        State0 = log(RelativeState0,[],[]),
        !.

% keeps the list of initial fact ids (for debugging purposes)
:- dynamic(initial_state_list/1).

% flush_initial_state/1
flush_initial_state(ListFactIds):-
	retractall(initial_state_list(_)),
	sort(ListFactIds,OrdListFactIds),
	assert(initial_state_list(OrdListFactIds)),
	create_initialState(ListFactIds).

% create_initialState/1
% create_initialState(+ListFactIds)
create_initialState([]).
create_initialState([H|T]):-
        assert(initial_state(H)),
        create_initialState(T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% absolute_state/2
absolute_state(log(S1,[],[]),S1):-
        !.
absolute_state(log(S1,LI1,LD1),S1t):-
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
        trie_interned(RelStateIns,InsDelsRoot,InsLeaf,_),
        trie_interned(RelStateDels,InsDelsRoot,DelsLeaf,_),
        mergeLogs(RelStateIns,RelStateDels,LI1,LD1,OrdIns,OrdDels),
        (
        	tr_debug(on)
        	-> inc_counter(state_comps,_)
        	; true
        ),
        trie_intern(OrdIns,InsDelsRoot,NewInsLeaf,_),
        trie_intern(OrdDels,InsDelsRoot,NewDelsLeaf,_),
	(
		(tr_debug(on),
		\+(trie_interned(state(NewInsLeaf,NewDelsLeaf),StateRoot)))
		-> inc_counter(state_copies,_)
		; true
	),
        trie_intern(state(NewInsLeaf,NewDelsLeaf),StateRoot,S1t,_),
        !.
absolute_state(S1,S1):-
        number(S1),
        !.

% differential_state/2
differential_state(log(S1,LI1,LD1),log(S1,LI1,LD1)):-
        !.
differential_state(S1,log(S1,[],[])):-
        number(S1),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_select_state
db_select_state(Fact,S1):-
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
	trie_interned(RelIns1,InsDelsRoot,InsLeaf,_),
	trie_interned(RelDels1,InsDelsRoot,DelsLeaf,_),
        fact_to_factid(Fact,FactId),
        db_select_state_internal(FactId,RelIns1,RelDels1).

% db_select_state_internal
db_select_state_internal(FactId,_RelIns1,RelDels1):-
        % the fact was in the initial state and was not deleted
        initial_state(FactId),
        \+( member_ord(FactId,RelDels1) ),
        !.
db_select_state_internal(FactId,RelIns1,_RelDels1):-
        % the fact was inserted in the relative state
        member_ord(FactId,RelIns1),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_select(+Fact,+RelativeState)
%checks if a fact is in the state
db_select(Fact,log(S1,LI1,LD1)):-
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
	trie_interned(RelIns1,InsDelsRoot,InsLeaf,_),
	trie_interned(RelDels1,InsDelsRoot,DelsLeaf,_),
        fact_to_factid(Fact,FactId),
        db_select_internal(FactId,RelIns1,RelDels1,LI1,LD1).

% db_select_internal/5
db_select_internal(FactId,RelIns1,RelDels1,_LI1,LD1):-
        % the fact was in the relative state and was not deleted
        db_select_state_internal(FactId,RelIns1,RelDels1),
        \+( member_ord(FactId,LD1) ),
        !.
db_select_internal(FactId,_RelIns1,_RelDels1,LI1,_LD1):-
        % the fact was inserted in the local log
        member_ord(FactId,LI1),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_ins(+Fact,
%   log(+RelativeState,+LI1,+LD1),
%   log(+RelativeState,-LI2,-LD2))
%  the tries for local logs LI1, LD1 get modified
db_ins(Fact,log(S1,LI1,LD1),log(S1,LI2,LD2)):-
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
	trie_interned(RelIns1,InsDelsRoot,InsLeaf,_),
	trie_interned(RelDels1,InsDelsRoot,DelsLeaf,_),
        no_backtracking_fact_to_factid(Fact,FactId),
        db_ins_internal(FactId,RelIns1,RelDels1,LI1,LD1,LI2,LD2).

% db_ins_internal(+FactId,+RelIns1,+RelDels1,+LI1,+LD1,-LI2,-LD2)
db_ins_internal(FactId,RelIns1,RelDels1,LI1,LD1,LI2,LD2):-
        % The fact is not in the relative state
        \+( db_select_state_internal(FactId,RelIns1,RelDels1) ),
        !,
        % if it is inserted in LI1,
        % then true (non-strict updates)
        % else insert it in LI1
        (
                member_ord(FactId,LI1)
                -> (LI2=LI1,LD2=LD1)
                ; (
                        ins_ord(FactId,LI1,LI2),
                        LD2=LD1
                )
        ).
db_ins_internal(FactId,_RelIns1,_RelDels1,LI1,LD1,LI2,LD2):-
        % The fact is in the relative state
        !,
        % if it is in the deleted trie LD1,
        % then delete it from LD1
        % else  true (non-strict updates)
        (
                member_ord(FactId,LD1)
                -> (
                        LI2=LI1,
                        del_ord(FactId,LD1,LD2)
                )
                ; (LI2=LI1,LD2=LD1)
        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_del(+Fact,
%   log(+RelaLIveState,+LI1,+LD1),
%   log(+RelaLIveState,-LI2,-LD2))
%  the tries for local logs LI1, LD1 get modified
db_del(Fact,log(S1,LI1,LD1),log(S1,LI2,LD2)):-
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
	trie_interned(RelIns1,InsDelsRoot,InsLeaf,_),
	trie_interned(RelDels1,InsDelsRoot,DelsLeaf,_),
        no_backtracking_fact_to_factid(Fact,FactId),
        db_del_internal(FactId,RelIns1,RelDels1,LI1,LD1,LI2,LD2).

% db_del_internal(+FactId,+RelIns1,+RelDels1,+LI1,+LD1,-LI2,-LD2)
db_del_internal(FactId,RelIns1,RelDels1,LI1,LD1,LI2,LD2):-
        % The fact is in the relaLIve state
        db_select_state_internal(FactId,RelIns1,RelDels1),
        !,
        % if it is inserted in LD1,
        % then true (non-strict updates)
        % else insert it in LD1
        (
                member_ord(FactId,LD1)
                -> (LI2=LI1,LD2=LD1)
                ; (
                        LI2=LI1,
                        ins_ord(FactId,LD1,LD2)
                )
        ).
db_del_internal(FactId,_RelIns1,_RelDels1,LI1,LD1,LI2,LD2):-
        % The fact is not in the relative state
        !,
        % if it is in the deleted trie LI1,
        % then delete it from LI1
        % else  true (non-strict updates)
        (
                member_ord(FactId,LI1)
                -> (
                        del_ord(FactId,LI1,LI2),
                        LD2=LD1
                )
                ; (LI2=LI1,LD2=LD1)
        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
