%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:-import new_trie/1,
        trie_intern/2,
        trie_intern/4,
        trie_interned/2,
        trie_interned/4 from intern.

:-import member/2,
        length/2,
        append/3,
        select/3 from basics.

:- import sort/2 from setof.

:- ['tr_utils.P'].

% dynamic initial_state/1 predicate: "initial_state(FactId)"
:- dynamic(initial_state/1).

% dynamic insDelsRoot/1 predicate containing the root of the trie that
%  contains materialized state logs in the format:
%  "ins/dels(Size,ListOfAll_FactIds_Ins/Dels_Ordered)"
:- dynamic(insDelsRoot/1).

% dynamic stateRoot/1 predicate contains the root of the trie that contains 
%  the states: "state(InsLeaf,DelsLeaf)"
:- dynamic(stateRoot/1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% initializeCTRSystem(+CTRProgram): initializes the transactional system
initializeCTRSystem(CTRProgram,State0):-
        retractall(dynamic_predicate(_,_)),
        retractall(tr_database(_)),
        retractall(tr_clause(_,_)),
        load_dyn(CTRProgram),
        findall(Head,tr_database(Head),ListFacts),
        no_backtracking_facts_to_factids(ListFacts,ListFactIds),
        retractall(initial_state(_)),
        retractall(insDelsRoot(_)),
        retractall(stateRoot(_)),
        flush_initial_state(ListFactIds),
        new_trie(InsDelsRoot),
        assert(insDelsRoot(InsDelsRoot)),
        new_trie(StateRoot),
        assert(stateRoot(StateRoot)),
        % saving the first state
        trie_intern([],InsDelsRoot,InsLeaf,_),
        trie_intern([],InsDelsRoot,DelsLeaf,_),
        trie_intern(state(InsLeaf,DelsLeaf),StateRoot,RelativeState0,_),
        (
        	tr_debug(on)
        	-> inc_counter(state_copies,_) % the initial state is created
        	; true
        ),
        new_trie(TI1),
        new_trie(TD1),
        State0 = log(RelativeState0,TI1,TD1),
        !.

% keeps the list of initial fact ids (for debugging purposes)
:- dynamic(initial_state_list/1).

% flush_initial_state/1
flush_initial_state(ListFactIds):-
	retractall(initial_state_list(_)),
	sort(ListFactIds,OrdListFactIds),
	assert(initial_state_list(OrdListFactIds)),
	create_initialState(ListFactIds).

% create_initialState/1
% create_initialState(+ListFactIds)
create_initialState([]).
create_initialState([H|T]):-
        assert(initial_state(H)),
        create_initialState(T).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% absolute_state/2
absolute_state(log(_S1,TI1,TD1),S1t):-
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_to_list(TI1,LI1U),
        sort(LI1U,LI1),
        trie_to_list(TD1,LD1U),
        sort(LD1U,LD1),
	(
        	tr_debug(on)
        	-> inc_counter(state_comps,_)
        	; true
        ),
        trie_intern(LI1,InsDelsRoot,NewInsLeaf,_),
        trie_intern(LD1,InsDelsRoot,NewDelsLeaf,_),
        (
		(tr_debug(on),
		\+(trie_interned(state(NewInsLeaf,NewDelsLeaf),
			StateRoot)))
		-> inc_counter(state_copies,_)
		; true
	),
        trie_intern(state(NewInsLeaf,NewDelsLeaf),StateRoot,S1t,_),
	!.
absolute_state(S1,S1):-
        number(S1),
        !.

% differential_state/2
differential_state(log(S1,TI1,TD1),log(S1,TI1,TD1)):-
        !.
differential_state(S1,log(S1,TI1,TD1)):-
        number(S1),
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
        trie_interned(LI1,InsDelsRoot,InsLeaf,_),
        trie_interned(LD1,InsDelsRoot,DelsLeaf,_),
        (
        	tr_debug(on)
        	-> (
        		inc_counter(state_copies,_)
        	); true
        ),
        list_to_trie(LI1,TI1),
        list_to_trie(LD1,TD1),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_select_state
db_select_state(Fact,S1):-
        stateRoot(StateRoot),
        insDelsRoot(InsDelsRoot),
        trie_interned(state(InsLeaf,DelsLeaf),StateRoot,S1,_),
	trie_interned(RelIns1,InsDelsRoot,InsLeaf,_),
	trie_interned(RelDels1,InsDelsRoot,DelsLeaf,_),
        fact_to_factid(Fact,FactId),
        db_select_state_internal(FactId,RelIns1,RelDels1).

% db_select_state_internal
db_select_state_internal(FactId,_RelIns1,RelDels1):-
        % the fact was in the initial state and was not deleted
        initial_state(FactId),
        \+( member_ord(FactId,RelDels1) ),
        !.
db_select_state_internal(FactId,RelIns1,_RelDels1):-
        % the fact was inserted in the relative state
        member_ord(FactId,RelIns1),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_select(+Fact,+RelativeState)
%checks if a fact is in the state
db_select(Fact,log(_S1,TI1,TD1)):-
        fact_to_factid(Fact,FactId),
        db_select_internal(FactId,TI1,TD1).

% db_select_internal/5
db_select_internal(FactId,_TI1,TD1):-
        % the fact was in the relative state and was not deleted
        initial_state(FactId),
        \+( trie_interned(FactId,TD1) ),
        !.
db_select_internal(FactId,TI1,_TD1):-
        % the fact was inserted in the local log
        trie_interned(FactId,TI1),
        !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_ins(+Fact,
%   log(+RelativeState,+TI1,+TD1),
%   log(+RelativeState,-TI1,-TD1))
%  the tries for local logs TI1, TD1 get modified
db_ins(Fact,log(S1,TI1,TD1),log(S1,TI1,TD1)):-
        no_backtracking_fact_to_factid(Fact,FactId),
        db_ins_internal(FactId,TI1,TD1).

% db_ins_internal(+FactId,+TI1,+TD1)
db_ins_internal(FactId,TI1,_TD1):-
        % The fact is not in the initial state
        \+( initial_state(FactId) ),
        !,
        % if it is inserted in TI1,
        % then true (non-strict updates)
        % else insert it in TI1
        (
                trie_interned(FactId,TI1)
                -> true
                ; bk_trie_intern(FactId,TI1)
        ).
db_ins_internal(FactId,_TI1,TD1):-
        % The fact is in the relative state
        !,
        % if it is in the deleted trie TD1,
        % then delete it from TD1
        % else  true (non-strict updates)
        (
                trie_interned(FactId,TD1,Leaf,_)
                -> bk_trie_unintern(FactId,TD1,Leaf)
                ; true
        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% db_del(+Fact,
%   log(+RelaTIveState,+TI1,+TD1),
%   log(+RelaTIveState,-TI1,-TD1))
%  the tries for local logs TI1, TD1 get modified
db_del(Fact,log(S1,TI1,TD1),log(S1,TI1,TD1)):-
        no_backtracking_fact_to_factid(Fact,FactId),
        db_del_internal(FactId,TI1,TD1).

% db_del_internal(+FactId,+TI1,+TD1)
db_del_internal(FactId,_TI1,TD1):-
        % The fact is in the initial state
        initial_state(FactId),
        !,
        % if it is inserted in TD1,
        % then true (non-strict updates)
        % else insert it in TD1
        (
                trie_interned(FactId,TD1)
                -> true
                ; bk_trie_intern(FactId,TD1)
        ).
db_del_internal(FactId,TI1,_TD1):-
        % The fact is not in the initial state
        !,
        % if it is in the deleted trie TI1,
        % then delete it from TI1
        % else  true (non-strict updates)
        (
                trie_interned(FactId,TI1,Leaf,_)
                -> bk_trie_unintern(FactId,TI1,Leaf)
                ; true
        ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
